@@ -1,11 +1,12 @@
 /* GDA Postgres provider
- * Copyright (C) 1998-2002 The GNOME Foundation
+ * Copyright (C) 1998-2005 The GNOME Foundation
  *
  * AUTHORS:
  *      Michael Lausch <michael@lausch.at>
  *	Rodrigo Moya <rodrigo@gnome-db.org>
  *      Vivien Malerba <malerba@gnome-db.org>
  *      Gonzalo Paniagua Javier <gonzalo@gnome-db.org>
+ *	Bas Driessen <bas.driessen@xobas.com>
  *
  * This Library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public License as
@@ -56,6 +57,8 @@ static const GdaValue *gda_postgres_recordset_get_value_at    (GdaDataModel *mod
 static GdaFieldAttributes *gda_postgres_recordset_describe    (GdaDataModel *model, gint col);
 static gint gda_postgres_recordset_get_n_rows 		      (GdaDataModel *model);
 static const GdaRow *gda_postgres_recordset_get_row 	      (GdaDataModel *model, gint rownum);
+static const GdaRow *gda_postgres_recordset_append_row        (GdaDataModel *model, const GList *values);
+static gboolean gda_postgres_recordset_remove_row 	      (GdaDataModel *model, const GdaRow *row);
 static gboolean gda_postgres_recordset_update_row 	      (GdaDataModel *model, const GdaRow *row);
 
 static GObjectClass *parent_class = NULL;
@@ -89,6 +92,8 @@ gda_postgres_recordset_class_init (GdaPostgresRecordsetClass *klass)
 	model_class->describe_column = gda_postgres_recordset_describe;
 	model_class->get_value_at = gda_postgres_recordset_get_value_at;
 	model_class->get_row = gda_postgres_recordset_get_row;
+	model_class->append_row = gda_postgres_recordset_append_row;
+	model_class->remove_row = gda_postgres_recordset_remove_row;
 	model_class->update_row = gda_postgres_recordset_update_row;
 }
 
@@ -155,6 +160,7 @@ get_row (GdaDataModel *model, GdaPostgresRecordsetPrivate *priv, gint rownum)
 
 	gda_row_set_number (row, rownum);
 	id = g_strdup_printf ("%d", rownum);
+   	/* Use oid or figure out primary keys ? could use libsql to add oid to every query. */
 	gda_row_set_id (row, id); /* FIXME: by now, the rowid is just the row number */
 	g_free (id);
 	return row;
@@ -174,7 +180,7 @@ gda_postgres_recordset_get_row (GdaDataModel *model, gint row)
 	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
 	g_return_val_if_fail (recset->priv != NULL, 0);
 
-	row_list = (GdaRow *) gda_data_model_hash_get_row (model, row);
+	row_list = (GdaRow *) GDA_DATA_MODEL_CLASS (parent_class)->get_row (model, row);
 	if (row_list != NULL)
 		return (const GdaRow *)row_list;
 
@@ -201,14 +207,231 @@ gda_postgres_recordset_get_row (GdaDataModel *model, gint row)
 	return (const GdaRow *) row_list;
 }
 
+static const GdaRow *
+gda_postgres_recordset_append_row (GdaDataModel *model, const GList *values)
+{
+	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
+	GdaPostgresRecordsetPrivate *priv_data;
+	GString *sql;
+	gint cols, i;
+	GList *l;
+	PGresult *pg_res;
+	GdaPostgresConnectionData *cnc_priv_data;
+	PGconn *pg_conn;
+	const GdaRow *row;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
+	g_return_val_if_fail (values != NULL, NULL);
+	g_return_val_if_fail (gda_data_model_is_updatable (model), NULL);
+	//g_return_val_if_fail (gda_data_model_hash_changed (model), NULL);
+	g_return_val_if_fail (recset->priv != NULL, NULL);
+	priv_data = recset->priv;
+	pg_res = priv_data->pg_res;
+	cnc_priv_data = g_object_get_data (G_OBJECT (priv_data->cnc),
+					   OBJECT_DATA_POSTGRES_HANDLE);
+	pg_conn = cnc_priv_data->pconn;
+
+	/* checks if the table name has been guessed */
+	if (priv_data->table_name == NULL) {
+		gda_connection_add_error_string (priv_data->cnc,
+						 _("Table name could not be guessed."));
+		return NULL;
+	}
+
+	cols = gda_data_model_get_n_columns (model);
+	if (cols != g_list_length ((GList *) values)) {
+		gda_connection_add_error_string (
+			priv_data->cnc,
+			_("Attempt to insert a row with an invalid number of columns"));
+		return NULL;
+	}
+
+	/* Prepare the SQL command */
+	sql = g_string_new ("INSERT INTO ");
+	g_string_append_printf (sql, "%s(",
+				priv_data->table_name);
+	for (i = 0; i < cols; i++) {
+		GdaFieldAttributes *fa;
+
+		fa = gda_data_model_describe_column (model, i);
+		if (!fa) {
+			gda_connection_add_error_string (
+				priv_data->cnc,
+				_("Could not retrieve column's information"));
+			g_string_free (sql, TRUE);
+			return NULL;
+		}
+
+		if (i != 0)
+			sql = g_string_append (sql, ", ");
+		sql = g_string_append (sql, "\"");
+		sql = g_string_append (sql, gda_field_attributes_get_name (fa));
+		sql = g_string_append (sql, "\"");
+	}
+	sql = g_string_append (sql, ") VALUES (");
+
+	for (l = (GList *) values, i = 0; i < cols; i++, l = g_list_next (l)) {
+		gchar *val_str;
+		const GdaValue *val = (const GdaValue *) l->data;
+	
+		if (!val) {
+			gda_connection_add_error_string (
+				priv_data->cnc,
+				_("Could not retrieve column's value"));
+			g_string_free (sql, TRUE);
+			return NULL;
+		}
+
+		if (i != 0)
+			sql = g_string_append (sql, ", ");
+		val_str = gda_postgres_value_to_sql_string ((GdaValue *)val);
+		if (val_str == NULL) {
+			gda_connection_add_error_string (
+				priv_data->cnc,
+				_("Could not retrieve column's value"));
+			g_string_free (sql, TRUE);
+			return NULL;
+		}
+		sql = g_string_append (sql, val_str);
+		g_free (val_str);
+	}
+	sql = g_string_append (sql, ")");
+
+	/* execute the SQL query */
+	pg_res = PQexec (pg_conn, sql->str);
+	g_string_free (sql, TRUE);
+
+	if (pg_res != NULL) {
+		/* update ok! */
+		if (PQresultStatus (pg_res) != PGRES_COMMAND_OK) {
+			gda_connection_add_error (priv_data->cnc,
+						  gda_postgres_make_error (pg_conn, pg_res));
+			PQclear (pg_res);
+			return NULL;
+		}
+		PQclear (pg_res);
+	}
+	else
+		gda_connection_add_error (priv_data->cnc,
+					  gda_postgres_make_error (pg_conn, NULL));
+
+	/* append row in hash table */
+	row = GDA_DATA_MODEL_CLASS (parent_class)->append_row (model, values);
+
+	return row;
+}
+
+static gboolean
+gda_postgres_recordset_remove_row (GdaDataModel *model, const GdaRow *row)
+{
+	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
+	GdaPostgresRecordsetPrivate *priv_data;
+	gint colnum, uk, i;
+	PGresult *pg_res, *pg_rm_res;
+	gchar *query, *query_where, *tmp;
+	GdaPostgresConnectionData *cnc_priv_data;
+	PGconn *pg_conn;
+	gboolean status = FALSE;
+	GdaValue *value;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), FALSE);
+	g_return_val_if_fail (recset->priv != NULL, FALSE);
+	g_return_val_if_fail (row != NULL, FALSE);
+
+	priv_data = recset->priv;
+	pg_res = priv_data->pg_res;
+	cnc_priv_data = g_object_get_data (G_OBJECT (priv_data->cnc),
+					   OBJECT_DATA_POSTGRES_HANDLE);
+	pg_conn = cnc_priv_data->pconn;
+
+	/* checks if the given row belongs to the given model */
+	if (gda_row_get_model ((GdaRow *) row) != GDA_DATA_MODEL (model)) {
+		gda_connection_add_error_string (priv_data->cnc,
+						_("Given row doesn't belong to the model."));
+		return FALSE;
+	}
+
+	/* checks if the table name has been guessed */
+	if (priv_data->table_name == NULL) {
+		gda_connection_add_error_string (priv_data->cnc,
+						_("Table name could not be guessed."));
+		return FALSE;
+	}
+
+	query_where = g_strdup ("WHERE TRUE ");
+
+	for (colnum = uk = 0;
+	     colnum != gda_data_model_get_n_columns (GDA_DATA_MODEL(model));
+	     colnum++)
+	{
+		GdaFieldAttributes *attrs = gda_data_model_describe_column (GDA_DATA_MODEL(model), colnum);
+		const gchar *column_name = PQfname (pg_res, colnum);
+		gchar *curval = gda_value_stringify (gda_row_get_value ((GdaRow *) row, colnum));
+
+		/* unique column: we will use it as an index */
+		if (gda_field_attributes_get_primary_key (attrs) ||
+			gda_field_attributes_get_unique_key (attrs))
+		{
+			/* fills the 'where' part of the update command */
+			tmp = g_strdup_printf ("AND %s = '%s' ",
+					       column_name,
+					       curval);
+			query_where = g_strconcat (query_where, tmp, NULL);
+			g_free (tmp);
+			uk++;
+		}
+
+		g_free (curval);
+		gda_field_attributes_free (attrs);
+	}
+
+	if (uk == 0) {
+		gda_connection_add_error_string (priv_data->cnc,
+						_("Model doesn't have at least one unique key."));
+	}
+	else {
+		/* build the delete command */
+		query = g_strdup_printf ("DELETE FROM %s %s",
+					 priv_data->table_name,
+					 query_where);
+
+		/* remove the row */
+		pg_rm_res = PQexec (pg_conn, query);
+		g_free (query);
+
+		if (pg_rm_res != NULL) {
+		/* removal ok! */
+			if (PQresultStatus (pg_rm_res) == PGRES_COMMAND_OK)
+				status = TRUE;
+			else
+				gda_connection_add_error (priv_data->cnc,
+							  gda_postgres_make_error (pg_conn, pg_rm_res));
+			PQclear (pg_rm_res);
+		}
+		else
+			gda_connection_add_error (priv_data->cnc,
+						  gda_postgres_make_error (pg_conn, NULL));
+	}
+
+	g_free (query_where);
+
+	/* remove entry from data model */
+	if (status == TRUE)
+		status = GDA_DATA_MODEL_CLASS (parent_class)->remove_row (model, row);
+
+	return status;
+}
+
 static gboolean
 gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 {
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 	GdaPostgresRecordsetPrivate *priv_data;
-	gint colnum, uk, nuk;
+	gint colnum, rownum, uk, nuk;
 	PGresult *pg_res, *pg_upd_res;
 	gchar *query, *query_where, *query_set, *tmp;
+	gchar *oldval, *newval;
+	const gchar *column_name;
 	GdaPostgresConnectionData *cnc_priv_data;
 	PGconn *pg_conn;
 	gboolean status = FALSE;
@@ -236,6 +459,8 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 		return FALSE;
 	}
 
+	rownum = gda_row_get_number ((GdaRow *) row);
+
 	query_where = g_strdup ("WHERE TRUE ");
 	query_set = g_strdup ("SET ");
 
@@ -244,9 +469,15 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 	     colnum++) 
 	{
 		GdaFieldAttributes *attrs = gda_data_model_describe_column (model, colnum);
-		const gchar *column_name = PQfname (pg_res, colnum);
-		gchar *newval = gda_value_stringify (gda_row_get_value ((GdaRow *) row, colnum));
-		const gchar *oldval = PQgetvalue (pg_res, gda_row_get_number ((GdaRow *) row), colnum);
+		column_name = PQfname (pg_res, colnum);
+		newval = gda_value_stringify (gda_row_get_value ((GdaRow *) row, colnum));
+
+		/* for data from mysql result we can retrieve original values to avoid
+		   unique columns to be updated */
+		if (rownum < priv_data->nrows)
+			oldval = PQgetvalue (pg_res, gda_row_get_number ((GdaRow *) row), colnum);
+		else
+			oldval = newval;
 
 		/* unique column: we won't update it, but we will use it as
 		   an index */
@@ -270,7 +501,7 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 		/* non-unique column: update it */
 		else {
 			/* fills the 'set' part of the update command */
-			tmp = g_strdup_printf ("%s = '%s', ", 
+			tmp = g_strdup_printf ("\"%s\" = '%s', ", 
 					       column_name,
 					       newval);
 			query_set = g_strconcat (query_set, tmp, NULL);
@@ -322,6 +553,11 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 	
 	g_free (query_set);
 	g_free (query_where);
+
+	/* emit update signals */
+	gda_data_model_row_updated (GDA_DATA_MODEL (model), gda_row_get_number ((GdaRow *) row));
+	gda_data_model_changed (GDA_DATA_MODEL (model));
+
 	return status;
 }
 
@@ -381,13 +617,17 @@ static gchar *guess_table_name (GdaPostgresRecordset *recset)
 	PGconn *pg_conn;
 	gchar *table_name = NULL;
 	
+   
+   	/* This code should be changed to use libsql, its proberly faster
+	 * in long run to store a parsed sql statement and then
+	 * just grab the tablename from that. */
 	pg_res = recset->priv->pg_res;
 	cnc_priv_data = g_object_get_data (G_OBJECT (recset->priv->cnc),
 					   OBJECT_DATA_POSTGRES_HANDLE);
 	pg_conn = cnc_priv_data->pconn;
 
 	if (PQnfields (pg_res) > 0) {
-		gchar *query = g_strdup_printf ("SELECT c.relname FROM pg_catalog.pg_class c WHERE c.relkind = 'r'");
+		gchar *query = g_strdup_printf ("SELECT c.relname FROM pg_catalog.pg_class c WHERE c.relkind = 'r' AND c.relnatts = '%d'", PQnfields (pg_res));
 		gint i;
 		for (i = 0; i < PQnfields (pg_res); i++) {
 			const gchar *column_name = PQfname (pg_res, i);
@@ -521,12 +761,19 @@ gda_postgres_recordset_describe (GdaDataModel *model, gint col)
 static gint
 gda_postgres_recordset_get_n_rows (GdaDataModel *model)
 {
+	gint parent_row_num;
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 
 	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (model), 0);
 	g_return_val_if_fail (recset->priv != NULL, 0);
 
-	return recset->priv->nrows;
+	parent_row_num = GDA_DATA_MODEL_CLASS (parent_class)->get_n_rows (model);
+
+	/* if not initialized return number of PQ Tuples */
+	if (parent_row_num < 0)
+		return recset->priv->nrows;
+	else
+		return parent_row_num;
 }
 
 /*
